@@ -6,31 +6,35 @@
 #include <time.h>
 
 #define MAX_BARCODE 1000
-#define MAX_HAIRPIN 10000
-#define SEQ_LEN 100
+#define MAX_HAIRPIN 100000
 #define BLOCKSIZE 10000000
 
 typedef struct {
    char   *sequence;
+   char   *sequenceRev;
    int    original_pos;
 } a_barcode;
 
 typedef struct {
    char   *sequence;
    int    original_pos;
-   long   count;
 } a_hairpin;
 
 a_barcode *barcodes[MAX_BARCODE];
 a_hairpin *hairpins[MAX_HAIRPIN];
 
+int isPairedReads;
 int num_barcode;
 int num_hairpin;
 long num_read;
+long hairpinreadcount[MAX_HAIRPIN];
 long summary[MAX_HAIRPIN][MAX_BARCODE];
 int barcode_start;
 int barcode_end;
+int barcode_start_rev;
+int barcode_end_rev;
 int barcode_length;
+int barcode_length_rev;
 int hairpin_start;
 int hairpin_end;
 int hairpin_length;
@@ -48,8 +52,6 @@ long hairpincount;
 long bchpcount;
 
 a_hairpin *mismatch_hairpins[MAX_HAIRPIN];
-int *barcodeindex;
-int *hairpinindex;
 
 void
 Read_In_Barcodes(char* filename){
@@ -63,18 +65,28 @@ Read_In_Barcodes(char* filename){
   a_barcode *new_barcode;
 
   int count = 0;
+  char * token;
+
   while ((readline = fgets(line, len, fin)) != NULL){
     count++;
     new_barcode = (a_barcode *)malloc(sizeof(a_barcode));
-    new_barcode->sequence = (char *)malloc(SEQ_LEN * sizeof(char));
-
-    new_barcode->original_pos = count;
+    new_barcode->sequence = (char *)malloc(barcode_length * sizeof(char));
     strncpy(new_barcode->sequence, line, barcode_length);
+    new_barcode->original_pos = count;
+    if (isPairedReads > 0) {
+      token = strtok(line, "\t");
+      token = strtok(NULL, "\t");
+      new_barcode->sequenceRev = (char *)malloc(barcode_length_rev * sizeof(char));
+      strncpy(new_barcode->sequenceRev, token, barcode_length_rev);
+    } else {
+      new_barcode->sequenceRev = NULL;
+    };
     barcodes[count] = new_barcode;
   }
   fclose(fin);
   num_barcode = count;
   free(line);
+
   Rprintf(" -- Number of Barcodes : %d\n", num_barcode);
 }
 
@@ -87,18 +99,47 @@ locate_barcode(char *a_barcode){
   while (imax >= imin) {
     imid = (imax + imin) / 2;
     
-    if (strncmp(barcodes[imid]->sequence, a_barcode, barcode_length) < 0)
+    if (strncmp(barcodes[imid]->sequence, a_barcode, barcode_length) < 0) {
       imin = imid + 1;
-    else if (strncmp(barcodes[imid]->sequence, a_barcode, barcode_length) > 0)
+    } else if (strncmp(barcodes[imid]->sequence, a_barcode, barcode_length) > 0) {
       imax = imid - 1;
-    else
+    } else {
       return barcodes[imid]->original_pos;
+    }
   }
   return -1;  
 }
 
 int
-locate_hairpin(char *a_hairpin){
+locate_barcode_paired(char *a_barcode, char *a_barcode_rev){
+  int imin, imax, imid;
+  imin = 1;
+  imax = num_barcode;
+
+  while (imax >= imin) {
+    imid = (imax + imin) / 2;
+    // compare forward barcode sequence
+    if (strncmp(barcodes[imid]->sequence, a_barcode, barcode_length) < 0) {
+      imin = imid + 1;
+    } else if (strncmp(barcodes[imid]->sequence, a_barcode, barcode_length) > 0) {
+      imax = imid - 1;
+    } else {
+      // same forward sequence, compare reverse barcode sequence
+      if (strncmp(barcodes[imid]->sequenceRev, a_barcode_rev, barcode_length_rev) < 0) {
+        imin = imid + 1;
+      } else if (strncmp(barcodes[imid]->sequenceRev, a_barcode_rev, barcode_length_rev) > 0) {
+        imax = imid - 1;
+      } else {
+        return barcodes[imid]->original_pos;     
+      } 
+    }    
+  }
+  return -1; 
+}
+
+
+int
+locate_hairpin_impl(char *a_hairpin){
   int imin, imax, imid;
   imin = 1;
   imax = num_hairpin;
@@ -106,12 +147,13 @@ locate_hairpin(char *a_hairpin){
   while (imax >= imin) {
     imid = (imax + imin) / 2;
     
-    if (strncmp(hairpins[imid]->sequence, a_hairpin, hairpin_length) < 0)
+    if (strncmp(hairpins[imid]->sequence, a_hairpin, hairpin_length) < 0) {
       imin = imid + 1;
-    else if (strncmp(hairpins[imid]->sequence, a_hairpin, hairpin_length) > 0)
+    } else if (strncmp(hairpins[imid]->sequence, a_hairpin, hairpin_length) > 0) {
       imax = imid - 1;
-    else
+    } else {
       return hairpins[imid]->original_pos;
+    }
   }
   return -1;  
 }
@@ -121,13 +163,15 @@ Valid_Match(char *sequence1, char *sequence2, int length, int threshold){
   int i_base;
   int mismatchbasecount = 0;
   for (i_base = 0; i_base < length; i_base++) {
-    if (sequence1[i_base] != sequence2[i_base])
+    if (sequence1[i_base] != sequence2[i_base]) {
       mismatchbasecount++;		  
+    }
   }
-  if (mismatchbasecount <= threshold)
+  if (mismatchbasecount <= threshold) {
     return 1;
-  else
+  } else {
     return -1;
+  }
 }
 
 int
@@ -143,6 +187,21 @@ locate_mismatch_barcode(char *a_barcode){
   return match_index;
 }
 
+int
+locate_mismatch_barcode_paired(char *a_barcode, char *a_barcode_rev){
+  int i;
+  int match_index = -1;
+  for (i = 1; i <= num_barcode; i++){
+    if ((Valid_Match(a_barcode, barcodes[i]->sequence, barcode_length, barcode_n_mismatch) > 0) && 
+	(Valid_Match(a_barcode_rev, barcodes[i]->sequenceRev, barcode_length_rev, barcode_n_mismatch) > 0)) {
+      match_index = barcodes[i]->original_pos;
+      break;
+    }
+  }
+  return match_index;
+}
+
+
 int
 locate_mismatch_hairpin(char *a_hairpin){
   int i;
@@ -157,13 +216,75 @@ locate_mismatch_hairpin(char *a_hairpin){
 }
 
 
+int 
+locate_hairpin(char *a_hairpin, char *read, int doMismatch){
+  int hairpin_index;
+  if (doMismatch > 0){
+    hairpin_index = locate_mismatch_hairpin(a_hairpin);
+  } else {
+    hairpin_index = locate_hairpin_impl(a_hairpin);
+  }
+
+  if ((hairpin_index <= 0) && (allow_shifting > 0)) {
+    if ((doMismatch > 0) && (allow_shifted_mismatch <= 0)) {
+      return hairpin_index;
+      exit(0); 
+    }
+    // Check if given hairpin can be mapped to a shifted location. 
+    char *shiftedharipinstr;
+    shiftedharipinstr = (char *)malloc(hairpin_length * sizeof(char));
+
+    int index;
+    // check shifting leftwards
+    for (index = 1; index <= shifting_n_base; index++){
+      strncpy(shiftedharipinstr, read + hairpin_start - 1 - index, hairpin_length);
+      if (doMismatch > 0){
+        hairpin_index = locate_mismatch_hairpin(shiftedharipinstr);
+      } else {
+        hairpin_index = locate_hairpin_impl(shiftedharipinstr);
+      }
+      if (hairpin_index > 0) {
+        break;
+      }
+    }
+    // check shifting rightwards
+    if (hairpin_index <= 0){
+      for (index = 1; index <= shifting_n_base; index++){
+        strncpy(shiftedharipinstr, read + hairpin_start - 1 + index, hairpin_length);
+        
+        if (doMismatch > 0){
+          hairpin_index = locate_mismatch_hairpin(shiftedharipinstr);
+        } else {
+          hairpin_index = locate_hairpin_impl(shiftedharipinstr);
+        }
+
+        if (hairpin_index > 0) {
+          break;
+        }
+      }
+    }
+  } 
+  return hairpin_index;
+}
+
+int 
+barcode_compare(a_barcode *barcode1, a_barcode *barcode2){
+  int ans;
+  ans = strncmp(barcode1->sequence, barcode2->sequence, barcode_length);
+  if ((ans == 0) && (isPairedReads > 0)){  
+    ans = strncmp(barcode1->sequenceRev, barcode2->sequenceRev, barcode_length_rev);
+  }
+
+  return ans;
+}
+
 void
 Sort_Barcodes(void){
   int i, j;
   a_barcode *temp;
   for(i = 1; i < num_barcode; i++){
     for(j = i+1; j <= num_barcode; j++){
-      if (strcmp(barcodes[i]->sequence, barcodes[j]->sequence) > 0){
+      if (barcode_compare(barcodes[i], barcodes[j]) > 0) {
 	temp = barcodes[i];
 	barcodes[i] = barcodes[j];
 	barcodes[j] = temp;
@@ -187,15 +308,19 @@ Read_In_Hairpins(char *filename){
   while ((readline = fgets(line, len, fin)) != NULL){
     count++;
     new_hairpin = (a_hairpin *)malloc(sizeof(a_hairpin));
-    new_hairpin->sequence = (char *)malloc(SEQ_LEN * sizeof(char));
+    new_hairpin->sequence = (char *)malloc(hairpin_length * sizeof(char));
     new_hairpin->original_pos = count;
-    new_hairpin->count = 0;
     strncpy(new_hairpin->sequence, line, hairpin_length);
     hairpins[count] = new_hairpin;
   }
   fclose(fin);
   num_hairpin = count;
   free(line);
+
+  int i_hairpin;
+  for (i_hairpin = 1; i_hairpin <= num_hairpin; i_hairpin++){
+    hairpinreadcount[i_hairpin] = 0;
+  }
   Rprintf(" -- Number of Hairpins : %d\n", num_hairpin);
 }
 
@@ -205,7 +330,7 @@ Sort_Hairpins(void){
   a_hairpin *temp;
   for(i = 1; i < num_hairpin; i++){
     for(j = i+1; j <= num_hairpin; j++){
-      if (strcmp(hairpins[i]->sequence, hairpins[j]->sequence) > 0){
+      if (strncmp(hairpins[i]->sequence, hairpins[j]->sequence, hairpin_length) > 0){
 	temp = hairpins[i];
 	hairpins[i] = hairpins[j];
 	hairpins[j] = temp;
@@ -214,108 +339,109 @@ Sort_Hairpins(void){
   }
 }
 
-long Count_Reads(char *filename) {
-  FILE *freads = fopen(filename, "r");
-  char * line = NULL;
-  char *readline;
-  size_t len = 1000;
-  line = (char *)malloc(sizeof(char) * (len + 1));
-  if(freads == NULL){
-    fclose(freads);
-    return 0;
-  }
-  long line_count = 0;
-  while ((readline = fgets(line, len, freads)) != NULL){
-    line_count++;
-  }
-  fclose(freads);
-  free(line);
-  return line_count / 4;
-}
-
-
 void
-Process_Hairpin_Reads(char *filename){
-  FILE *fin;
+Process_Hairpin_Reads(char *filename, char *filename2){
+  FILE *fin = NULL;
+  FILE *finRev = NULL; 
   char * line = NULL;
+  char * line2 = NULL;
   size_t len = 1000;
   char *readline;
+  char *readline2;
   long num_read_thisfile = 0;
 
-  fin = fopen(filename,"r");
+  char *this_barcode_for = NULL;
+  char *this_barcode_rev = NULL;
+  char *this_hairpin = NULL;
+
   line = (char *)malloc(sizeof(char) * (len+1));
-  
-  if (isverbose)
-    Rprintf("Processing reads in %s.\n", filename);
+  fin = fopen(filename,"r");
+  if (isPairedReads > 0) {
+    finRev = fopen(filename2, "r");
+    line2 = (char *)malloc(sizeof(char) * (len+1));
+  }
 
-  char * this_barcode;
-  char * this_hairpin;
-  this_barcode = (char *)malloc(SEQ_LEN * sizeof(char));
-  this_hairpin = (char *)malloc(SEQ_LEN * sizeof(char));
+  if (isverbose > 0){
+    if (isPairedReads > 0) {
+      Rprintf("Processing reads in %s and %s.\n", filename, filename2);
+    } else {
+      Rprintf("Processing reads in %s.\n", filename);
+    }
+  }
 
+  this_barcode_for = (char *)malloc(barcode_length * sizeof(char));
+  if (isPairedReads > 0) {
+    this_barcode_rev = (char *)malloc(barcode_length_rev * sizeof(char));
+  }
+  this_hairpin = (char *)malloc(hairpin_length * sizeof(char));
   long line_count = 0;
 
   int barcode_index;
   int hairpin_index;
 
   while ((readline = fgets(line, len, fin)) != NULL){
-    line_count++;  
-    if ((line_count % 4) != 2)
-      continue;
-    
-    if ((isverbose) && (num_read_thisfile % BLOCKSIZE == 0))
-      Rprintf(" -- Processing %d million reads\n", (num_read_thisfile / BLOCKSIZE + 1) * 10);
-    num_read++;
-    num_read_thisfile++;
-    strncpy(this_barcode, line + barcode_start - 1, barcode_length);
-    barcode_index = locate_barcode(this_barcode);
-
-    strncpy(this_hairpin, line + hairpin_start - 1, hairpin_length);
-    hairpin_index = locate_hairpin(this_hairpin);
-
-    if ((hairpin_index <= 0) && (allow_shifting > 0)){
-      // Check if given hairpin can be mapped to a shifted location. 
-      int index;
-      // check shifting leftwards
-      for (index = 1; index <= shifting_n_base; index++){
-	strncpy(this_hairpin, line + hairpin_start - 1 - index, hairpin_length);
-	hairpin_index = locate_hairpin(this_hairpin);
-	if (hairpin_index > 0)
-	  break;
-      }
-      // check shifting rightwards
-      if (hairpin_index <= 0){
-	for (index = 1; index <= shifting_n_base; index++){
-	  strncpy(this_hairpin, line + hairpin_start - 1 + index, hairpin_length);
-	  hairpin_index = locate_hairpin(this_hairpin);
-	  if (hairpin_index > 0)
+    if (isPairedReads > 0) {
+      readline2 = fgets(line2, len, finRev);
+	  if (readline2 == NULL) {
 	    break;
-	}
-      }
+	  }
+    }
+    line_count++;  
+    if ((line_count % 4) != 2) {
+      continue;
     }
 
-    if (barcode_index > 0)
+    if ((isverbose > 0) && (num_read_thisfile % BLOCKSIZE == 0)) {
+      Rprintf(" -- Processing %d million reads\n", (num_read_thisfile / BLOCKSIZE + 1) * 10);
+    }
+    num_read++;
+    num_read_thisfile++;
+  
+    strncpy(this_barcode_for, line + barcode_start - 1, barcode_length);
+    if (isPairedReads > 0){    
+      strncpy(this_barcode_rev, line2 + barcode_start_rev - 1, barcode_length_rev);
+      barcode_index = locate_barcode_paired(this_barcode_for, this_barcode_rev); 
+    } else { 
+      barcode_index = locate_barcode(this_barcode_for);
+    }
+
+    strncpy(this_hairpin, line + hairpin_start - 1, hairpin_length);
+    hairpin_index = locate_hairpin(this_hairpin, line, 0); // not allowing mismatch
+
+    if (barcode_index > 0){
       barcodecount++;
+    }
 
     if (hairpin_index > 0){
       hairpincount++;
-      hairpins[hairpin_index]->count++;
+      hairpinreadcount[hairpin_index]++;
     }
        
     if ((barcode_index > 0) && (hairpin_index > 0)) {
       summary[hairpin_index][barcode_index]++;
       bchpcount++;
     }
-    
-    barcodeindex[num_read] = barcode_index;
-    hairpinindex[num_read] = hairpin_index;
+
+  } // end while
+
+  if (isverbose > 0) {
+    if (isPairedReads > 0) {
+      Rprintf("Number of reads in file %s and %s: %ld\n", filename, filename2, num_read_thisfile);  
+    } else {
+      Rprintf("Number of reads in file %s : %ld\n", filename, num_read_thisfile);  
+    }
   }
-  if (isverbose)
-    Rprintf("Number of reads in file %s : %ld\n", filename, num_read_thisfile);  
-  fclose(fin);
+ 
+  fclose(fin); 
   free(line);
-  free(this_barcode);
-  free(this_hairpin);  
+  free(this_barcode_for);
+  free(this_hairpin); 
+
+  if (isPairedReads > 0){  
+    fclose(finRev);
+    free(line2);
+    free(this_barcode_rev);
+  }
 }
 
 void
@@ -324,7 +450,7 @@ Create_Mismatch_Hairpins_List(void){
   num_mismatch_hairpin = 0;
 
   for (i = 1; i <= num_hairpin; i++){
-    if (hairpins[i]->count == 0){
+    if (hairpinreadcount[i] == 0){
       num_mismatch_hairpin++;
       mismatch_hairpins[num_mismatch_hairpin] = hairpins[i];
     }
@@ -334,98 +460,132 @@ Create_Mismatch_Hairpins_List(void){
 
 
 void
-Process_Mismatch(char *filename){
+Process_Mismatch(char *filename, char *filename2){
 
-  FILE *fin;
+  FILE *fin = NULL;
+  FILE *finRev = NULL;
   char * line = NULL;
+  char * line2 = NULL;
   size_t len = 1000;
   char *readline;
+  char *readline2;
   long num_read_thisfile = 0;
 
-  fin = fopen(filename,"r");
-  line = (char *)malloc(len+1);
-
-  if (isverbose)
-    Rprintf("Processing reads in %s, considering sequence mismatch. \n", filename);
-
   char * this_hairpin;
-  char * this_barcode;
-  this_hairpin = (char *)malloc(SEQ_LEN * sizeof(char));
-  this_barcode = (char *)malloc(SEQ_LEN * sizeof(char));
+  char * this_barcode_for;
+  char * this_barcode_rev = NULL;
+
+  line = (char *)malloc(len+1);
+  fin = fopen(filename,"r");
+  if (isPairedReads > 0) {
+    finRev = fopen(filename2, "r");
+    line2 = (char *)malloc(sizeof(char) * (len+1));
+  }
+  if (isverbose > 0){
+    if (isPairedReads > 0) {
+      Rprintf("Re-processing reads in %s and %s, considering sequence mismatch\n", filename, filename2);
+    } else {
+      Rprintf("Re-processing reads in %s, considering sequence mismatch\n", filename);
+    }
+  }
+
+  this_barcode_for = (char *)malloc(barcode_length * sizeof(char));
+  if (isPairedReads > 0) {
+    this_barcode_rev = (char *)malloc(barcode_length_rev * sizeof(char));
+  }
+  this_hairpin = (char *)malloc(hairpin_length * sizeof(char));
 
   long line_count = 0;
 
+  int barcode_index;
+  int hairpin_index;
+
   int new_barcode_index;
   int new_hairpin_index;
-  while ((readline = fgets(line, len, fin)) != NULL){
-    line_count++;
-    if ((line_count % 4) != 2)
-      continue;
 
-    if ((isverbose) && (num_read_thisfile % BLOCKSIZE == 0))
+  while ((readline = fgets(line, len, fin)) != NULL){
+    if (isPairedReads > 0) {
+      readline2 = fgets(line2, len, finRev);
+	  if (readline2 == NULL) {
+	    break;
+	  }
+    }
+
+    line_count++;
+    if ((line_count % 4) != 2) {
+      continue;
+    }
+
+    if ((isverbose > 0) && (num_read_thisfile % BLOCKSIZE == 0)) {
       Rprintf(" -- Processing %d million reads\n", (num_read_thisfile / BLOCKSIZE + 1) * 10);
+    }
     num_read++;
     num_read_thisfile++;
 
-    // only re-process reads withougt perfect hairpin match or without perfect barcode match;
-    if ((hairpinindex[num_read] > 0) && (barcodeindex[num_read] > 0))
-      continue;
-
-    // re-match barcode:
-    if (barcodeindex[num_read] <= 0){
-      strncpy(this_barcode, line + barcode_start - 1, barcode_length);
-      new_barcode_index = locate_mismatch_barcode(this_barcode);
-      if (new_barcode_index > 0)
-	barcodecount++;
-    } else {
-        new_barcode_index = barcodeindex[num_read];
+    // re-do the mapping in Process_Hairpin_Reads()
+    strncpy(this_barcode_for, line + barcode_start - 1, barcode_length);
+    if (isPairedReads > 0){    
+      strncpy(this_barcode_rev, line2 + barcode_start_rev - 1, barcode_length_rev);
+      barcode_index = locate_barcode_paired(this_barcode_for, this_barcode_rev); 
+    } else { 
+      barcode_index = locate_barcode(this_barcode_for);
     }
- 
-    // re-match hairpin:
-    if (hairpinindex[num_read] <= 0){
-    	strncpy(this_hairpin, line + hairpin_start - 1, hairpin_length);
-    	new_hairpin_index = locate_mismatch_hairpin(this_hairpin);
-	if ((new_hairpin_index <= 0) && (allow_shifting > 0) && (allow_shifted_mismatch > 0)){
-	// Check if given hairpin can be mapped to a shifted location. 
-	  int index;
-	  // check shifting leftwards
-	  for (index = 1; index <= shifting_n_base; index++){
-	    strncpy(this_hairpin, line + hairpin_start - 1 - index, hairpin_length);
-	    new_hairpin_index = locate_mismatch_hairpin(this_hairpin);
-	    if (new_hairpin_index > 0)
-	      break;
-	  }
-	  // check shifting rightwards
-	  if (new_hairpin_index <= 0){
-	    for (index = 1; index <= shifting_n_base; index++){
-	      strncpy(this_hairpin, line + hairpin_start - 1 + index, hairpin_length);
-	      new_hairpin_index = locate_mismatch_hairpin(this_hairpin);
-	      if (new_hairpin_index > 0)
-		break;
-	    }
-	  }
-	}
-	if (new_hairpin_index > 0)
-	  hairpincount++;
-        } else {
-    	  new_hairpin_index = hairpinindex[num_read];
-        }
 
-      if ((new_barcode_index > 0) && (new_hairpin_index > 0)) {
-        summary[new_hairpin_index][new_barcode_index]++;
-        bchpcount++;
+    strncpy(this_hairpin, line + hairpin_start - 1, hairpin_length);
+    hairpin_index = locate_hairpin(this_hairpin, line, 0); //not allowing mismatch
+
+    // only re-process reads without perfect hairpin match or without perfect barcode match;    
+    if ((barcode_index > 0) && (hairpin_index > 0)) {
+      continue;
+    }
+
+    if (barcode_index > 0) {
+      new_barcode_index = barcode_index;
+    } else {
+      if (isPairedReads > 0){
+        new_barcode_index = locate_mismatch_barcode_paired(this_barcode_for, this_barcode_rev);
+      } else {
+        new_barcode_index = locate_mismatch_barcode(this_barcode_for);
       }
 
-  }
-  fclose(fin);
+      if (new_barcode_index > 0) {
+	barcodecount++;
+      }
+    } 
+ 
+    // re-match hairpin:
+    if (hairpin_index > 0){
+      new_hairpin_index = hairpin_index;
+    } else {
+      new_hairpin_index = locate_hairpin(this_hairpin, line, 1);
+
+      if (new_hairpin_index > 0) {
+        hairpincount++;
+      } 
+    }
+  
+    if ((new_barcode_index > 0) && (new_hairpin_index > 0)) {
+      summary[new_hairpin_index][new_barcode_index]++;
+      bchpcount++;
+    }
+  } // end while 
+
+  fclose(fin); 
   free(line);
-  free(this_barcode);
-  free(this_hairpin);
+  free(this_barcode_for);
+  free(this_hairpin); 
+
+  if (isPairedReads > 0){  
+    fclose(finRev);
+    free(line2);
+    free(this_barcode_rev);
+  }
+
 }
 
 
 void 
-Initialise(int barcodestart, int barcodeend, int hairpinstart, int hairpinend, 
+Initialise(int IsPaired, int barcodestart, int barcodeend, int barcodestartrev, int barcodeendrev, int hairpinstart, int hairpinend, 
 	   int allowshifting, int shiftingnbase,
 	   int allowMismatch, int barcodemismatch, int hairpinmismatch, 
 	   int allowShiftedMismatch, int verbose){
@@ -438,11 +598,15 @@ Initialise(int barcodestart, int barcodeend, int hairpinstart, int hairpinend,
   num_barcode = 0;
   num_hairpin = 0;
 
+  isPairedReads = IsPaired;
   barcode_start = barcodestart;
   barcode_end = barcodeend;
+  barcode_start_rev = barcodestartrev;
+  barcode_end_rev = barcodeendrev;
   hairpin_start = hairpinstart;
   hairpin_end = hairpinend;
   barcode_length = barcode_end - barcode_start + 1;
+  barcode_length_rev = barcode_end_rev - barcode_start_rev + 1;
   hairpin_length = hairpin_end - hairpin_start + 1;
 
   allow_shifting = allowshifting;
@@ -493,6 +657,9 @@ Clean_Up(void){
   int index;
   for (index = 1; index <= num_barcode; index++){
     free(barcodes[index]->sequence);
+    if (isPairedReads > 0){
+      free(barcodes[index]->sequenceRev);
+    }
     free(barcodes[index]);
   }
   for (index = 1; index <= num_hairpin; index++){
@@ -501,38 +668,35 @@ Clean_Up(void){
   }
 }
 
+
 void 
-processHairpinReads(char **file, int *filecount, 
+processHairpinReads(int *isPairedReads,
+                    char **file, char **file2, int *filecount, 
 		    char**barcodeseqs, char**hairpinseqs, 
-		    int *barcodestart, int *barcodeend, int *hairpinstart, int *hairpinend, 
+		    int *barcodestart, int *barcodeend, int *barcodestartrev, int *barcodeendrev, int *hairpinstart, int *hairpinend, 
 		    int *allowShifting, int *shiftingnbase, 
 		    int *allowMismatch, int *barcodemismatch, int *hairpinmismatch,
 		    int *allowShiftedMismatch,
 		    char **output, int *verbose)
 {  
-  Initialise(*barcodestart, *barcodeend, *hairpinstart, *hairpinend, 
+  Initialise(*isPairedReads, *barcodestart, *barcodeend, *barcodestartrev, *barcodeendrev, *hairpinstart, *hairpinend, 
 	     *allowShifting, *shiftingnbase,
 	     *allowMismatch, *barcodemismatch, *hairpinmismatch, 
 	     *allowShiftedMismatch, *verbose);
 
   Read_In_Barcodes(*barcodeseqs);
-  Sort_Barcodes();
+  Sort_Barcodes(); // () is necessary to invoke function without parameters
 
   Read_In_Hairpins(*hairpinseqs);
-  Check_Hairpins(); 
-  Sort_Hairpins();
 
-  long totalreads;
+  Check_Hairpins();
+  Sort_Hairpins();  
+  
   int i_file;
-  totalreads = 0;
+
   for (i_file = 0; i_file < *filecount; i_file++){
-    totalreads = totalreads + Count_Reads(file[i_file]);
+    Process_Hairpin_Reads(file[i_file], file2[i_file]);
   }
-  barcodeindex = (int *)malloc(sizeof(int) * totalreads);
-  hairpinindex = (int *)malloc(sizeof(int) * totalreads);
-  for (i_file = 0; i_file < *filecount; i_file++){
-    Process_Hairpin_Reads(file[i_file]);
- }
  
   if (allow_mismatch > 0){
     num_read = 0; 
@@ -540,13 +704,20 @@ processHairpinReads(char **file, int *filecount,
     Create_Mismatch_Hairpins_List();
     if (num_mismatch_hairpin > 0){
       for (i_file = 0; i_file < *filecount; i_file++){
-	Process_Mismatch(file[i_file]);
+	if (isPairedReads > 0) {
+          Process_Mismatch(file[i_file], file2[i_file]);
+        } else {
+          Process_Mismatch(file[i_file], NULL);
+        }
       }
     }  
   }
 
   Rprintf("\nThe input run parameters are: \n");
   Rprintf(" -- Barcode: start position %d\t end position %d\t length %d\n", barcode_start, barcode_end, barcode_length);  
+  if (isPairedReads > 0){
+    Rprintf(" -- Barcode in reverse read: start position %d\t end position %d\t length %d\n", barcode_start_rev, barcode_end_rev, barcode_length_rev); 
+  }
   Rprintf(" -- Hairpin: start position %d\t end position %d\t length %d\n", hairpin_start, hairpin_end, hairpin_length); 
   if (allow_shifting) {
     Rprintf(" -- Allow hairpin sequences to be matched to a shifted position, <= %d base left or right of the specified positions. \n", shifting_n_base);
@@ -557,7 +728,7 @@ processHairpinReads(char **file, int *filecount,
     Rprintf(" -- Allow sequence mismatch, <= %d base in barcode sequence and <= %d base in hairpin sequence. \n", barcode_n_mismatch, hairpin_n_mismatch );
   } else {
     Rprintf(" -- Mismatch in barcode/hairpin sequences not allowed. \n");
-  }
+  } 
 
   Rprintf("\nTotal number of read is %ld \n", num_read);
   Rprintf("There are %ld reads (%.4f percent) with barcode matches\n", barcodecount, 100.0*barcodecount/num_read);
@@ -565,7 +736,13 @@ processHairpinReads(char **file, int *filecount,
   Rprintf("There are %ld reads (%.4f percent) with both barcode and hairpin matches\n", bchpcount, 100.0*bchpcount/num_read);
 
   Output_Summary_Table(*output);
-  free(barcodeindex);
-  free(hairpinindex);
+
+  Clean_Up();
 }
 
+
+
+
+
+
+       
